/******************************************************************************
MODULE: huffman.cpp
PROEJCT: newhuffman
DATE: October 1997

USAGE: huffman <frequency_filename>
******************************************************************************/
#include <iostream.h>
#include <iomanip.h>
#include <stdio.h>
#include <fstream.h>
#include <string.h>
#ifdef UNIX
#include <strstream.h>
#else
#include <strstrea.h>
#endif
#include <assert.h>

#ifndef SWAN
#define SWAN
#endif

#include "sail.h"
#include "slist.h"
#include "Huffman.h"

template <class ELEM>
void BinNode<ELEM>::draw(GRAPHID gid)
//
//Purpose:		Recursively draw a Huffman tree node and its descendents in Swan.
//Parameters:	gid - a Swan graphic object id
//Return:		void
//
{
	char str[10];
	ostrstream ostr(str, 10);

	if (this->isLeaf()) {
		ostr.seekp(0);
		ostr << *this << ends;
		sw_insertnode(gid, NODEID(this), str);
		sw_setnodeattr(gid, NODEID(this), GNODELINETH, THICKLINE);
		sw_setnodeattr(gid, NODEID(this), GNODEWIDTH, 46);
	}
	else {
		ostr.seekp(0);
		ostr << *this << ends;
		sw_insertnode(gid, NODEID(this), str);
		sw_setnodeattr(gid, NODEID(this), GNODETYPE, CIRCLE);
		sw_setnodeattr(gid, NODEID(this), GNODEWIDTH, 35);
		left->draw(gid);
		sw_insertbinedge(gid, NODEID(this), NODEID(left), "0", LEFTCHILD);
		right->draw(gid);
		sw_insertbinedge(gid, NODEID(this), NODEID(right), "1", RIGHTCHILD);
	}
}

HuffDemo::HuffDemo(void)
{
	tmplist = new list<HuffTree*>;
	finalTree = NULL;
	n_graphs = 0;
}

void HuffDemo::clear(void)
//
//Purpose:		Clear all data connected to a HuffDemo structure.
//Parameters:	void
//Return:		void
//
{
	HuffTree* ht;

	tmplist->setFirst();

	while(ht = tmplist->currValue()) {
		delete ht;
		tmplist->next();
	}

	tmplist->clear();
	delete tmplist;

	if (finalTree) {
		finalTree->release(finalTree->root());
		delete finalTree;
	}

	erase_graphs();
}

bool HuffDemo::init(ifstream fs)
//
// Purpose:		Creates a partial Huffman tree list from an input data file.
// Parameters:	fs - a file stream attached to an input file
// Return:		true  - the initialisation success
//				false - failure
{
	LettFreq lf;
	HuffTree *ht, *htemp;
	int count = 0;

	while (1) {
		fs >> lf.lett >> lf.freq;
		if (fs.eof() || fs.fail()) {
			if (count == 0)		// fail if there's no data
				return false;
			else				// succeed
			{
				build_graphs(); // create the initial graphs for the Huffman tree list
				return true;
			}
		}

		ht = new HuffTree(lf);	// create a partial tree node

		tmplist->setFirst();

		if (htemp = tmplist->currValue())			// insert the new tree into the sorted list
			while (ht->weight() > htemp->weight()) {	// by its root node weight
				tmplist->next();
				if (!(htemp = tmplist->currValue())) break;
			}

		if (!tmplist->insert(ht))	return false;	// fail i
		
		count++;
	}
}

void HuffDemo::build_graphs(void)
//
// Purpose:		Build Swan graphs from the initial partial Huffman tree list.
// Parameters:	void
// Return:		void
//
{
	HuffTree* ht;
	
	tmplist->setFirst();

	while (ht = tmplist->currValue()) {
		GRAPHID gid = sw_newgraph(GRAPHID(++n_graphs), BOX, UNDIRECTED);
		LC_ID lc = sw_newlc(gid, NULLLCID, TREE);
		sw_setcurlc(gid, lc);
		ht->draw(gid);
		tmplist->next();
	}
	sw_displayallgraphs();
}

void HuffDemo::erase_graphs(void)
//
// Purpose:		Clear all graphs.
// Parameters:	void
// Return:		void
//
{
	for (int i = 1; i <= n_graphs; i++)
		sw_deletegraph(GRAPHID(i));
	n_graphs = 0;
}

void HuffDemo::build_tree(void)
//
// This code mainly comes from Dr. Shaffer's book 'Data Structures and Algorithm Analysis'.
//
// Purpose:		Build Huffman tree from a partial Huffman tree list.
// Parameters:	void
// Return:		void
//
{
	HuffTree *temp1, *temp2, *temp3;
	LettFreq* tempnode;

	for (tmplist->setPos(1); tmplist->isInList(); tmplist->setPos(1)) {
		tmplist->setFirst();		// if there are more than one node in the list
		temp1 = tmplist->remove();	// emerge the first two nodes into a new Huffman tree
		temp2 = tmplist->remove();	// whose weight is 
		tempnode = new LettFreq(temp1->weight() + temp2->weight());
		temp3 = new HuffTree(*tempnode, temp1, temp2);
		delete tempnode;
		delete temp1;
		delete temp2;

		for (tmplist->setFirst(); tmplist->isInList(); tmplist->next())
			if (temp3->weight() <= tmplist->currValue()->weight()) {
				tmplist->insert(temp3);	// insert the new Huffman tree back to the list
				break;
			}
		if (!tmplist->isInList())
			tmplist->append(temp3);
		erase_graphs();		// display the changes by Swan
		build_graphs();
		sw_print(" The two trees with lowest frequency have been combined. Press STEP to continue");
		sw_break();
	}

	tmplist->setFirst();
	finalTree = tmplist->remove();
	build_codetable(finalTree->root(),"");	// build code table for the Huffman tree
}

void HuffDemo::build_codetable(BinNode<LettFreq>* node, char* code)
//
// Purpose:		Recursively build a code table from a Huffman tree.
// Parameters:	node - a Huffman tree node
//				code - the Huffman code for the node
// Return:		void
//
{
	char tmp_str[MAX_CODE_LENGTH];

	if (node->leftchild() == NULL && node->rightchild() == NULL) {
		codetable.set(node->value().letter(), code);
		return;		// if the node is a leaf node, then add it to the code table
	}

	strcpy(tmp_str, code);	// build code table for its left child
	strcat(tmp_str, "0");
	build_codetable(node->leftchild(), tmp_str);

	strcpy(tmp_str, code);	// build code table for its right child
	strcat(tmp_str, "1");
	build_codetable(node->rightchild(), tmp_str);
}

void HuffDemo::test_code(void)
//
// Purpose:		Receive user inputs and do encoding and decoding demo.
// Parameters:	void
// Return:		void
//
{
	char input_buf[MAX_STR_LENGTH];		// user input buffer
	char out_buf[MAX_STR_LENGTH];		// output buffer
	char result_buf[MAX_STR_LENGTH];	// result buffer
	char *word_ptr;		// pointer to the user input 'word' position
	char *str;			// pointer to the current input buffer position
	char *result;		// pointer to the current result buffer position
	bool success;
	ostrstream ostr(out_buf, MAX_STR_LENGTH);	// string stream used for output

	while (1) {
		sw_step();
		sw_print("Enter \"e <word>\" to encode or \"d <bitstring>\" to decode:");
		sw_getstr(input_buf);
		str = input_buf;
		result = result_buf;

		switch(*str) {
		case 'd':		// do decoding
			success = true;
			str++;
			while(*str == ' ') str++;	// ignore the spaces between 'd' and codes
			word_ptr = str;

			while (*str != 0) {	// the pointer str will be modified in the function decode()
				if (!(*result = decode(finalTree->root(), str))) {
					success = false;
					break;
				}
				result++;
			}
			*result = 0;

			if (success) {
				ostr.seekp(0);
				ostr << "The word for <" << word_ptr << "> is <" << result_buf
					<< "> Press STEP or Quit." << ends;
				sw_print(out_buf);
			}
			else
				sw_print("Error in bitstring. Press STEP or Quit.");

			break;

		case 'e':		// do encoding
			success = true;
			str++;
			while(*str == ' ') str++;
			word_ptr = str;

			while (*str != 0) {
				if (!codetable.get(*str, result)) {
					success = false;
					break;
				}
				while (*result)	result++;
				str++;
			}

			if (success) {
				ostr.seekp(0);
				ostr << "The code for <" << word_ptr << "> is <" << result_buf
					<< "> Press STEP or Quit." << ends;
				sw_print(out_buf);
			}
			else
				sw_print("Error in word. Press STEP or Quit.");

			break;

		default:
			sw_print("Input error, press STEP to continue.");
		}

	sw_break();
	}
}

char HuffDemo::decode(BinNode<LettFreq>* node, char* &str)
//
// Purpose:		Recursively decode a bitstring
// Parameters:	node - a Huffman tree node
//				str - address of a pointer that points to the current position of a bitstring
// Return:		character decoded from the bitstring
//
{
	if (*str == 0)	{	// reach the end of the bitstring
		if ( node->leftchild() == NULL && node->rightchild() == NULL)
			return node->value().letter();	// if it's a leaf node, return the letter corresponding to it
		else			// if it's not a leaf node, the bitstring is wrong.
			return 0;
	}

	if (*str == '0') {
		if (node->leftchild() != NULL) {
			str++;
			return decode(node->leftchild(), str);
		}
		else
			return node->value().letter();
	}

	if (*str == '1') {
		if (node->rightchild() != NULL) {
			str++;
			return decode(node->rightchild(), str);
		}
		else
			return node->value().letter();
	}

	return 0;
}

//////////////////////////////////////////////////////////////////////////
//
// Huffman Demo main function.
//
//////////////////////////////////////////////////////////////////////////
int GS_MAIN(GS_MAIN_ARGS_F)
{
	char    fileName[80];
	HuffDemo demo;

	sw_init(GS_MAIN_ARGS_A);
	sw_step();
	sw_print("What file would you like to use?");
	sw_getstr(fileName);

	ifstream fs(fileName, ios::in | ios::nocreate);

	while (1) {
		while (fs.fail()) {
			sw_print("Couldn't open file, Press STEP to go on");
			sw_wait(RUN | STEP);
			sw_print("What file would you like to use?");
			sw_getstr(fileName);
			fs.open(fileName, ios::in | ios::nocreate);
		}	

		if (!demo.init(fs)) {	// build the initial partial Huffman tree list
			sw_print("Data init error, Press STEP to go on");
			fs.close();
			demo.clear();

			sw_wait(RUN | STEP);
			sw_print("What file would you like to use?");
			sw_getstr(fileName);
			fs.open(fileName, ios::in | ios::nocreate);
		}
		else
			break;
	}

	sw_print("Here are the leaf nodes. Press STEP to build the Huffman tree.");
	sw_break();
	demo.build_tree();	// demostrate Huffman tree building
	sw_print("This is the final Huffman tree. Press STEP to continue.");
	sw_break();
	demo.test_code();	// do encoding and decoding demo

	return 0;
}
