//
// Huffman.h
// This code mainly comes from Dr. Shaffer's book 'Data Structures and Algorithm Analysis'.
//

#define MAX_CODE_LENGTH	10
#define MAX_CODE_NUM	100
#define MAX_STR_LENGTH	256

class CodeTable {	// A code table holds the Huffman codes of a character set
	struct {
		char lett;					// a character
		char code[MAX_CODE_LENGTH];	// it's Huffman code
	}codes[MAX_CODE_NUM];

	int n_codes;		// number of codes(characters)
public:
	
	CodeTable(void) { n_codes = 0; }
	
	~CodeTable() { }
	
	void set(char c, char* code)	// Insert a character and its Huffman code into the code table
	{
	  int i;
		for (i = 0; i < n_codes; i++)
			if (codes[i].lett == c) break;

		if (i == n_codes) {
			n_codes++;
			codes[i].lett = c;
		}

		strcpy(codes[i].code, code);
	}

	bool get(char c, char* code)	// Retrieve the Huffman code of a character from the code table
	{
	  int i;
		for (i = 0; i < n_codes; i++)
			if (codes[i].lett == c) break;

		if (i == n_codes) return false;

		strcpy(code, codes[i].code);
		
		return true;
	}
};

class LettFreq {
public:
	char lett;
	int freq;

	LettFreq(int f = 0, char l = '\0') { freq = f; lett = l; }

	LettFreq(LettFreq& lf) { freq = lf.freq; lett = lf.lett; }

	~LettFreq() { }

	int weight(void) { return freq; }

	char letter(void) { return lett; }

	bool operator == (LettFreq& lf) { return freq == lf.freq; }
	
	bool operator > (LettFreq& lf)	{ return freq > lf.freq; }

	bool operator < (LettFreq& lf)	{ return freq < lf.freq; }

	friend ostream& operator<< (ostream& os, LettFreq& lf)
	{
		if (lf.lett)
			os << lf.lett << ":" << lf.freq << "\0";
		else
			os << lf.freq << "\0";
		return os;
	}
};

template <class ELEM>
class BinNode {
	ELEM element;
	BinNode* left;
	BinNode* right;
public:
	BinNode(void) { left = right = NULL; }

	BinNode(ELEM& e, BinNode<ELEM>* l = NULL, BinNode<ELEM>* r = NULL)
		{ element = e; left = l; right = r; }

	~BinNode() { }

	BinNode* leftchild(void) { return left; }

	BinNode* rightchild(void) { return right; }

	ELEM& value(void) { return element; }

	BinNode* setValue(ELEM& val) { element = val; return this;}

	bool isLeaf(void) { return (left == NULL) && (right == NULL); }

	void draw(GRAPHID);

	bool operator > (BinNode<ELEM>& b)	{ return element > b.element; }

	bool operator < (BinNode<ELEM>& b)	{ return element < b.element; }

	friend ostream& operator<< (ostream& os, BinNode<ELEM>& b)
	{ os << b.element; return os; }
};

class HuffTree {
	BinNode<LettFreq>* rt;
public:
	HuffTree(LettFreq& val) { rt = new BinNode<LettFreq>(val); }

	HuffTree(LettFreq& val, HuffTree* leftree, HuffTree* rightree)
		{ rt = new BinNode<LettFreq>(val, leftree->rt, rightree->rt); }

	~HuffTree() { }

	void release(BinNode<LettFreq>* node)
	{
		if (node == NULL) return;
		release(node->leftchild());
		release(node->rightchild());
		delete node;
	}

	void draw(GRAPHID gid)	{ rt->draw(gid); }
	
	bool isEmpty(void) const { return rt == NULL; }
	
	BinNode<LettFreq>* root(void) { return rt ; }
	
	int weight(void) { return rt->value().weight(); }
	
	bool operator > (HuffTree& ht)	{ return *rt > *ht.rt; }

	bool operator < (HuffTree& ht)	{ return *rt < *ht.rt; }
};

class HuffDemo {
	list<HuffTree*> *tmplist;	// a list of partial Huffman trees
	HuffTree* finalTree;	// a pointer to the buildup Huffman tree
	int n_graphs;			// 
	CodeTable codetable;	// a table to store the codes of letters for encoding
public:
	HuffDemo(void);

	~HuffDemo() { clear(); }

	void clear(void);

	bool init(ifstream);

	void build_tree(void);

	void build_graphs(void);

	void erase_graphs(void);

	void build_codetable(BinNode<LettFreq>*, char*);

	void test_code(void);

	char decode(BinNode<LettFreq>*, char*&);
};

//End of Huffman.h
