/*****************************************************************************
MODULE: kmpsearch.cpp
PROJECT: kmpsearch
AUTHOR: Jeff Nielsen
DATE: May 1996
DESCRIPTION: An implementation of the Knuth-Morris-Pratt string-searching
             algorithm, annotated with SAIL function calls.
*****************************************************************************/
#ifndef SWAN
#define SWAN
#endif
#include <iostream.h>
#include <fstream.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <strstream.h>
#include "stdjeff.h"

#define MAX_PATTERN_SIZE 20	 // Longest pattern that can be searched for

// Swan functions -------------------------------------------------------------
#ifdef SWAN
#include "sail.h"

// Names for Swan graphs in this visualization
GRAPHID searchmc = GRAPHID(100);       // The "search machine" graph
GRAPHID patttable = GRAPHID(101);      // The pattern array
GRAPHID nexttable = GRAPHID(102);      // The "next" array
GRAPHID textstream = GRAPHID(103);     // The array used to show the textstream
char memstr[300];    // Global memory location for building strings
strstream memstm(memstr, 300, ios::in | ios::out);  // Attach stream

//////////////////////////////////////////////////
inline void make_swangraph(GRAPHID gid, NODETYPE nt, bool dir_undir,
            LAYOUTCOMPTYPE lctype)
//
// PURPOSE: Macro for making SWAN graphs
// PARAMETERS:
//    gid: The graphid you want for this graph
//    nt: The type of node you want the graph to have
//    dir_undir: DIRECTED or UNDIRECTED graph
//    lctype: The default layout component you want this graph to have.
{
   sw_newgraph(gid, nt, dir_undir);
   sw_newlc(gid, LC_ID(1), lctype);
   sw_setcurlc(gid, LC_ID(1));
}

//////////////////////////////////////////////////
void draw_graphs(char *pat)
//
// PURPOSE: Draw all of the graphs for this visualization in the Swan window
// PARAMETER:
//    pat: The string pattern to be searched for
{
   // Create the "searching machine," pattern table, and "next table" graphs
   int i, patlength = strlen(pat);
   make_swangraph(searchmc, CIRCLE, DIRECTED, CIRCLENET);
   sw_setlcattr(searchmc, LC_ID(1), GNODEWIDTH, 35);
   make_swangraph(patttable, BOX, UNDIRECTED, ARRAYDOWN);
   sw_setgraphattr(patttable, GLAYOUTMODE, MANULAYOUT);
   sw_setgraphpos(patttable, 449, 25);
   make_swangraph(nexttable, BOX, UNDIRECTED, ARRAYDOWN);
   sw_setgraphattr(nexttable, GLAYOUTMODE, MANULAYOUT);
   sw_setgraphpos(nexttable, 475, 25);

   // Insert nodes in all three graphs
   sw_insertnode(searchmc, NODEID(-1), "-1");	   // First node in search m.c.
   for (i = 0; i <= patlength; i++) {
      if (i != patlength) {
	 memstm.seekp(0); memstm << pat[i] << ends;
	 sw_insertnode(searchmc, NODEID(i), memstr);
	 sw_insertnode(patttable, NODEID(i), memstr);
	 sw_insertnode(nexttable, NODEID(i), " ");
      }
      else {					   // Last node in search m.c.
	 sw_insertnode(searchmc, NODEID(i), "match");
	 sw_insertnode(nexttable, NODEID(i), " ");
      }
      sw_insertedge(searchmc, NODEID(i-1), NODEID(i), " ");
   }

   // Create the text stream graph
   make_swangraph(textstream, BOX, DIRECTED, ARRAYACROSS);
   sw_setgraphattr(textstream, GLAYOUTMODE, MANULAYOUT);
   sw_setgraphpos(textstream, 20, 380);
   sw_insertnode(textstream, NODEID(1), " ");
   sw_setnodeattr(textstream, NODEID(1), GNODEWIDTH, 123);
   sw_insertnode(textstream, NODEID(2), " ");
   sw_setnodeattr(textstream, NODEID(3), GNODEWIDTH, 5);
   sw_insertnode(textstream, NODEID(3), " ");
   sw_setnodeattr(textstream, NODEID(3), GNODEWIDTH, 245);

   sw_displayallgraphs();	 // Show all the graphs
}

//////////////////////////////////////////////////
void show_textstream(istream &is, int textpos)
//
// PURPOSE: Shows the current text stream in the textstream graph
// PARAMETERS:
//    is: The input (text) stream
//    textpos: The current offset in the textstream
{
   int readlength;

   // Show 20 "used" characters in node 1
   if (textpos >= 20) readlength = 20;
   else readlength = textpos;
   is.seekg(-1 * (readlength+1), ios::cur);
   is.read(memstr, readlength);
   memstr[readlength] = 0;
   sw_setnodeattr(textstream, NODEID(1), GNODELABEL, memstr);

   // Show the current character in node 2
   is.read(memstr, 1);
   memstr[1] = 0;
   sw_setnodeattr(textstream, NODEID(2), GNODELABEL, memstr);

   // Show 40 characters to come in node 3
   is.read(memstr, 40);
   if (!is) { readlength = is.gcount() ; is.clear(); }
   else readlength = 40;
   memstr[readlength] = 0;
   sw_setnodeattr(textstream, NODEID(3), GNODELABEL, memstr);
   is.seekg(-1 * (readlength), ios::cur);
}

//////////////////////////////////////////////////
// PURPOSE: Macro used twice in the initnext function.  Inserts an edge in the
//	    searchmc graph and a number in the nexttable graph, corresponding
//	    to the appropriate entry in the "next" table.
//
#define ADDEDGE \
   sw_insertedge(searchmc, NODEID(i), NODEID(next[i]), " "); \
   memstm.seekp(0); \
   memstm << next[i] << ends; \
   sw_setnodeattr(nexttable, NODEID(i), GNODELABEL, memstr); \
   if (i < m-1) { \
      sw_print("Building the \"next\" table.  Press STEP or RUN."); \
      sw_break(); \
   }

//////////////////////////////////////////////////
// PURPOSE: Macros used in the kmpsearch function.  MARK_NODE changes the
//	    color of the current node in the searchmc graph and fills it in.
//	    UNMARK_NODE changes the node back
//
#define MARK_NODE  sw_setnodeattr(searchmc, NODEID(j), GNODEFILLED, TRUE); \
                   sw_setnodeattr(searchmc, NODEID(j), GNODECOLOR, LYELLOW)
#define UNMARK_NODE sw_setnodeattr(searchmc, NODEID(j), GNODEFILLED, FALSE); \
		    sw_setnodeattr(searchmc, NODEID(j), GNODECOLOR, DGREEN)  


#endif

// KMP Algorithm functions ---------------------------------------------------

//////////////////////////////////////////////////
void report_match_at_position(int p)
//
// PURPOSE: Reports a match
// PARAMETER:
//    p: The position in the text at which the match was found
{ cout << "Match found at position " << p << endl; }

//////////////////////////////////////////////////
void initnext(char pat[], int m, int next[])
//
// PURPOSE: Initilalize the "next" table.  Based on code from the Frakes book,
//	    p. 223.
// PARAMETERS:
//    pat: Array containing the string pattern to search for
//    m: Length of the pattern array + 1
//    next: Array containing the table to initialize
{
   int i, j;
   i = 0; j = next[0] = -1;
#ifdef SWAN
   ADDEDGE;
#endif
   do {
      if (j == -1 || pat[i] == pat[j]) {
         i++; j++;
         if (pat[i] != pat[j]) next[i] = j;
         else next[i] = next[j];
#ifdef SWAN
	 ADDEDGE;
#endif
      }
      else j = next[j];
   } while (i < m-1);
}

//////////////////////////////////////////////////
void kmpsearch(istream &is, char pat[], int m)
//
// PURPOSE: Knuth-Morris-Pratt string-searching algorithm.  Based on code from
//	    the Frakes book, p. 223.
// PARAMETERS:
//    is: The input stream containing the text to be searched
//    pat: The array containing the string pattern to search for
//    m: The length of the string being searched for
{
   int j,                           // position in pattern
       k,                           // position in text
       resume;                      // where to start in pattern after a match
   char textchar;                   // current character in text
   int next[MAX_PATTERN_SIZE+1];    // holds "next" table

   // Initilialize everything  -------------------------------
   initnext(pat, m+1, next);        // Preprocess pattern
   resume = next[m];
   j = k = 0;                                                           
   is.get(textchar);                // Get first character in text
   if (!is) return;      			// No text to search
#ifdef SWAN
   sw_step();			 // Go back to STEP mode if they are in RUN
   MARK_NODE;			 // Identify the starting node
   show_textstream(is, k); // Show the textstream in the textstream graph
   sw_print("Table complete--ready to begin searching.  "
	    "Press STEP or RUN to continue.");
   sw_break();
   UNMARK_NODE;
#endif

   // Main search loop --------------------------------------
   while (TRUE) {                             // Exits on EOF
      if (j == -1 || textchar == pat[j] ) {
         is.get(textchar);             // Advance to next character in text
	 if (is.eof()) break;          // Exit loop on EOF
         k++; j++;                     // Advance to next character in pattern
#ifdef SWAN
	 if (j != 0) sw_print("Character matched--both pointers advanced.");
	 else sw_print("Reached \"reset\" state--both pointers advanced.");
#endif
      }
      else {
	 j = next[j];
#ifdef SWAN
	 sw_print("Mismatch--pattern pointer moved according to table.");
#endif
      }
      if (j == m) {		       // Match was found
#ifdef SWAN
	 MARK_NODE;
	 show_textstream(is, k);
	 sw_displayallgraphs();
	 memstm.seekp(0);
	 memstm << "Match found at text position " << k-m
		<< "--text pointer advanced."  << ends;
	 sw_print(memstr);
	 sw_step();	   // Stop when match found
	 sw_break();
	 UNMARK_NODE;
	 sw_print("Pattern pointer moved according to table.");
#else
	 report_match_at_position(k-m);
#endif
         j = resume;
      }	    // end if (j == m)
#ifdef SWAN
      MARK_NODE;
      show_textstream(is, k);
      sw_displayallgraphs();
      sw_break();
      UNMARK_NODE;
#endif
   }	    // end while (TRUE)
}

//////////////////////////////////////////////////
int GS_MAIN(GS_MAIN_ARGS_F)
//
// PURPOSE: Driver for program
// PARAMETERS: The standard ones
{
   char    temp_str[80] ;
   char pattern[MAX_PATTERN_SIZE];
   int readlength;

#ifdef SWAN
   sw_init(GS_MAIN_ARGS_A);

   sw_print("What file would you like to use?") ;
   sw_getstr(temp_str) ;
   ifstream inputfile(temp_str);
   if (!inputfile) {
      cout << "Couldn't open file " << temp_str << endl;
      exit(-1);
   }

   sw_step();
   inputfile.read(temp_str, 20);
   readlength = inputfile.gcount();
   if (readlength == 0) readlength = 20;
   inputfile.clear();
   temp_str[readlength] = 0;
   inputfile.seekg(-1 * readlength, ios::cur);
   sprintf(memstr, "Text string is |%s|. Enter a search pattern.",temp_str);
   sw_print(memstr);
   sw_getstr(memstr);
   sscanf(memstr, "%s", pattern);
   draw_graphs(pattern);		     // Set up all 4 graphs
#else
   cout << "Enter a pattern to search for: ";
   cin.get(pattern, MAX_PATTERN_SIZE);
#endif
   
   if (pattern[0] != 0) {
      kmpsearch(inputfile, pattern, strlen(pattern));
#ifdef SWAN
      sw_step();
      sw_print("EOF reached--no more matches.");
      sw_break();
#endif
   }
   else cout << "No pattern entered" << endl;
   return 0;
}

