/*****************************************************************************
MODULE: redbfunc.cxx
PROJECT: redblack
PROGRAMMER: Jeff Nielsen
DATE: February 1996

DESCRIPTION: Contains the function definitions for the Red-Black tree classes.
             Annotated with SAIL function calls to be able to visualize the
	     tree.  Compile with -DSWAN for the annotated version.

******************************************************************************/

// Header Files ---------------------------------------------------------------
#ifndef SWAN
#define SWAN
#endif
#include <iostream.h>
#include <strstream.h>
#include <iomanip.h>
#include <ctype.h>

#include "redblack.h"
#include "stdjeff.h"

#ifdef SWAN
#include "sail.h"
void rebuild_tree(redblack_tree*);     // Prototype for Swan drawing function
char memstr[100];		       // Global memory string space
strstream memstm(memstr, 100, ios::in | ios::out);
#endif

// rb_node class methods ----------------------------------------------------

// GLOBAL VARIABLES
rb_node* rb_node::freelist = NULL;  // Allocate space for the freelist ptr
rb_node* NULLNODE = new rb_node;    // Make a NULL rb_node colored BLCK

//////////////////////////////////////////////////
rb_node::rb_node(ELEM k, COLORTYPE c)
//
// PURPOSE: "Regular" constructor--sets children to point to NULLNODE
// PARAMETERS:
//    k: Value for new node
//    c: Color for new node--defaults to RED
   { key = k; color = c; left = right = NULLNODE; }

//////////////////////////////////////////////////
void* rb_node::operator new(size_t)
//
// PURPOSE: Overload the C++ new operator to allocate nodes from a freelist.
//          If no nodes on the freelist, use the system new operator.
// RETURNS: Pointer to allocated node
{
   if (freelist == NULL) return ::new rb_node;
   rb_node* temp = freelist;
   freelist = freelist->left;       // Grab the last node put on the freelist
   return temp;
}

//////////////////////////////////////////////////
void rb_node::operator delete(void* ptr)
//
// PURPOSE: Overload C++ delete operator to put returned nodes on a freelist.
{
   ((rb_node*)ptr)->left = freelist;  // Use the left ptr to point to next
   freelist = (rb_node*)ptr;          //    node on list
}


// redblack_tree class methods -------------------------------------------

/////////////////////////////////////////////////////////////////
redblack_tree::redblack_tree() { root = NULLNODE; }         // Constructor

/////////////////////////////////////////////////////////////////
void redblack_tree::clear()
//
// PURPOSE: Clear the tree
{
   clearhelp(root);
   root = NULLNODE;
   cout << "The tree has been emptied.";
}

/////////////////////////////////////////////////////////////////
bool redblack_tree::is_empty() const { return root == NULLNODE; }

/////////////////////////////////////////////////////////////////
void redblack_tree::insert(const ELEM& val)
//
// PURPOSE: Insert value in tree by calling recursive inserthelp() function
// PARAMETERS:
//    val: Value to be inserted
{
   if (val >= MIN && val <= MAX) inserthelp(root, val, NULLNODE);
   else {
#ifdef SWAN
      memstm.seekp(0);
      memstm << ERRSTREAM << ends;
      sw_print(memstr);
#else
      cout << ERRSTREAM;
#endif
   }
}

/////////////////////////////////////////////////////////////////
void redblack_tree::remove(const ELEM& val)
//
// PURPOSE: Remove value from tree
// PARAMETERS:
//    val: Value to remove
{
   if (val >= MIN && val <= MAX) removehelp(root, val);
   else cout << ERRSTREAM;
}

/////////////////////////////////////////////////////////////////
void redblack_tree::left_rotate(rb_node *x)
//
// PURPOSE: Perform a left rotation on two tree nodes.  This code
// comes from p.266 of Cormen, et al, Introduction to Algorithms.
//
// PARAMETERS:
// x: pointer to "parent" node to be rotated
{
#ifdef SWAN
   rebuild_tree(this);
   memstm.seekp(0);
   memstm << "Have to do a left rotation on node " << x->key
	  << ".  Press STEP to continue." << ends;
   sw_print(memstr);
   sw_break();
#endif
   rb_node* y = x->right;  // Set y
   x->right = y->left;     // Turn y's left subtree into x's right subtree
   if (y->left != NULLNODE) y->left->p = x;
   y->p = x->p;            // Link x's parent to y
   if (x->p == NULLNODE) root = y;   // Reset root pointer if necessary
   else if (x == x->p->left) x->p->left = y;
   else x->p->right = y;

   y->left = x;            // Put x on y's left
   x->p = y;
}

/////////////////////////////////////////////////////////////////
void redblack_tree::right_rotate(rb_node *x)
//
// PURPOSE: Perform a right rotation on two tree nodes.  This code
// comes from p.266 of Cormen, et al, Introduction to Algorithms.
//
// PARAMETERS:
// x: pointer to "parent" node to be rotated
{
#ifdef SWAN
   rebuild_tree(this);
   memstm.seekp(0);
   memstm << "Have to do a right rotation on node " << x->key
	  << ".  Press STEP to continue." << ends;
   sw_print(memstr);
   sw_break();
#endif
   rb_node* y = x->left;      // Set y
   x->left = y->right;        // Turn y's right subtree into x's left subtree
   if (y->right != NULLNODE) y->right->p = x;
   y->p = x->p;               // Link x's parent to y
   if (x->p == NULLNODE) root = y;   // Reset root pointer if necessary
   else if (x == x->p->right) x->p->right = y;
   else x->p->left = y;

   y->right = x;              // Put x on y's right
   x->p = y;
}


/////////////////////////////////////////////////////////
void redblack_tree::fix_tree(rb_node* x)
// PURPOSE: Adjust the tree so that red-black properties are preserved.
//          This code comes from p.268 of Cormen, et al.
// PARAMETERS:
//    x: Pointer to current node
{
   rb_node* y;
   while (x != root && x->p->color == RED) {
      if (x->p == x->p->p->left) {    // If x's parent is a left child
         y = x->p->p->right;                // Set y to be x's uncle
         if (y->color == RED) {
#ifdef SWAN
	 rebuild_tree(this);
	 sw_print("Insert Case 1--Press STEP to continue.");
	 sw_break();
#endif
            x->p->color = BLCK;
            y->color = BLCK;
            x->p->p->color = RED;
            x = x->p->p;               // Move x to point at grandparent
         }
         else {
            if (x == x->p->right) {
#ifdef SWAN
	 rebuild_tree(this);
	 sw_print("Insert Case 2--Press STEP to continue.");
	 sw_break();
#endif
               x = x->p;               // Case 2
               left_rotate(x);         // Case 2
            }
#ifdef SWAN
	 rebuild_tree(this);
	 sw_print("Insert Case 3--Press STEP to continue.");
	 sw_break();
#endif
            x->p->color = BLCK;        // Case 3
            x->p->p->color = RED;      // Case 3
            right_rotate(x->p->p);     // Case 3
         }
      }
      else {                        // x's parent is a right child
         y = x->p->p->left;            // Set y to be x's uncle
         if (y->color == RED) {
#ifdef SWAN
	 rebuild_tree(this);
	 sw_print("Insert Case 1--Press STEP to continue.");
	 sw_break();
#endif
            x->p->color = BLCK;
            y->color = BLCK;
            x->p->p->color = RED;
            x = x->p->p;            // Move x to point at grandparent
         }
         else {
            if (x == x->p->left) {
#ifdef SWAN
	 rebuild_tree(this);
	 sw_print("Insert Case 2--Press STEP to continue.");
	 sw_break();
#endif
               x = x->p;               // Case 2
               right_rotate(x);        // Case 2
            }
#ifdef SWAN
	 rebuild_tree(this);
	 sw_print("Insert Case 3--Press STEP to continue.");
	 sw_break();
#endif
            x->p->color = BLCK;       // Case 3
            x->p->p->color = RED;      // Case 3
            left_rotate(x->p->p);      // Case 3
         }
      }
   }
   if (root->color == RED) {
#ifdef SWAN
      rebuild_tree(this);
      sw_print("Now make root black.  Press STEP to continue.");
      sw_break();
#endif
      root->color = BLCK;
   }
}


//////////////////////////////////////////////////
void redblack_tree::inserthelp(rb_node*& rt, const ELEM val, rb_node* par)
//
// PURPOSE: Recursive function for inserting a value in the tree.  If current
//          node is empty makes a new node with the value.  Otherwise it calls
//          itself on the appropriate subtree.  After value is inserted
//          fix_tree() is called to restore the red-black properties.
// PARAMETERS:
//   rt: current node's pointer--ref. variable because it is changed
//   val: Value to be inserted
//   par: Parent of current node
{
   if (rt == NULLNODE) { // Node is empty--we're in the right place in tree
      rt = new rb_node(val);
      rt->p = par;
      fix_tree(rt);
#ifdef SWAN
      memstm.seekp(0);
      memstm << "Finished inserting value " << val << '.' << ends;
      sw_print(memstr);
#else
      cout << "Value " << val << " inserted.";
#endif      
   }

   else if (val == rt->key) {
#ifdef SWAN
      memstm.seekp(0);
      memstm << "ERROR: The tree already contains value " << val << '.' <<ends;
      sw_print(memstr);
#else      
      cout << "ERROR: The tree already contains value " << val << '.';
#endif
   }
   else if (val < rt->key) inserthelp(rt->left, val, rt);
   else inserthelp(rt->right, val, rt);

}

//////////////////////////////////////////////////
void redblack_tree::delete_fixup(rb_node* x)
//
// PURPOSE: Restore red-black properties after a delete operation.  From
//          Cormen, et al, p.274
// PARAMETERS:
//    x: Pointer to deleted node's sole child or to NULLNODE if deleted
//       node had no children
{
   rb_node* w;
   while (x != root && x->color == BLCK) {

      if (x == x->p->left) {
         w = x->p->right;        // w is x's sibling
         if (w->color == RED) {
#ifdef SWAN
	 rebuild_tree(this);
	 sw_print("Delete Case 1--Press STEP to continue.");
	 sw_break();
#endif
            w->color = BLCK;
            x->p->color = RED;
            left_rotate(x->p);
            w = x->p->right;
         }
         if (w->left->color == BLCK && w->right->color == BLCK) {
#ifdef SWAN
	 rebuild_tree(this);
	 sw_print("Delete Case 2--Press STEP to continue.");
	 sw_break();
#endif
            w->color = RED;
            x = x->p;
         }
         else {
            if (w->right->color == BLCK) {
#ifdef SWAN
	 rebuild_tree(this);
	 sw_print("Delete Case 3--Press STEP to continue.");
	 sw_break();
#endif
               w->left->color = BLCK;
               w->color = RED;
               right_rotate(w);
               w = x->p->right;
            }
#ifdef SWAN
	 rebuild_tree(this);
	 sw_print("Delete Case 4--Press STEP to continue.");
	 sw_break();
#endif
            w->color = x->p->color;
            x->p->color = BLCK;
            w->right->color = BLCK;
            left_rotate(x->p);
            x = root;
         }
      }
      else {
         w = x->p->left;         // w is x's sibling
#ifdef SWAN
	 rebuild_tree(this);
	 sw_print("Delete Case 1--Press STEP to continue.");
	 sw_break();
#endif
         if (w->color == RED) {
            w->color = BLCK;
            x->p->color = RED;
            right_rotate(x->p);
            w = x->p->left;
         }
         if (w->right->color == BLCK && w->left->color == BLCK) {
#ifdef SWAN
	 rebuild_tree(this);
	 sw_print("Delete Case 2--Press STEP to continue.");
	 sw_break();
#endif
            w->color = RED;
            x = x->p;
         }
         else {
            if (w->left->color == BLCK) {
#ifdef SWAN
	 rebuild_tree(this);
	 sw_print("Delete Case 3--Press STEP to continue.");
	 sw_break();
#endif
               w->right->color = BLCK;
               w->color = RED;
               left_rotate(w);
               w = x->p->left;
            }
#ifdef SWAN
	 rebuild_tree(this);
	 sw_print("Delete Case 4--Press STEP to continue.");
	 sw_break();
#endif
            w->color = x->p->color;
            x->p->color = BLCK;
            w->left->color = BLCK;
            right_rotate(x->p);
            x = root;
         }
      }
   }
   if (x->color == RED) {
#ifdef SWAN
      rebuild_tree(this);
      memstm.seekp(0);
      memstm << "Now make node " << x->key
	     << " black.  Press STEP to continue." << ends;
      sw_print(memstr);
      sw_break();
#endif
      x->color = BLCK;
   }
}

//////////////////////////////////////////////////
void redblack_tree::removehelp(rb_node* rt, const ELEM val)
//
// PURPOSE: Recursive function for removing a value from the tree.  Mostly
//          from Cormen, et al, p.273
// PARAMETERS:
//   rt: Pointer to current node
//   val: Value to be removed
{

   if (rt == NULLNODE)	// Don't need to check if using Swan
      cout << "ERROR: Value " << val << " is not in the tree.";
   else if (val < rt->key) removehelp(rt->left, val);
   else if (val > rt->key) removehelp(rt->right, val);
   else {                                 // Found it--now remove it
      rb_node *x, *y;

      if (rt->left == NULLNODE || rt->right == NULLNODE) y = rt;
      else {
	 y = findmin(rt->right);        // Successor of rt
#ifdef SWAN
	 memstm.seekp(0);
	 memstm << "Replacing node "<< rt->key << " with node "
		<< y->key << ".  Press STEP to continue." << ends;
	 sw_print(memstr);
	 sw_break();
#endif
      }
      if (y->left != NULLNODE) x = y->left;
      else x = y->right;

      x->p = y->p;
      if (y->p == NULLNODE) root = x;
      else if (y == y->p->left) y->p->left = x;
      else y->p->right = x;

      if (y != rt) rt->key = y->key;      // Move successor's value to rt

      if (y->color == BLCK) delete_fixup(x);
      delete y;
#ifdef SWAN
      memstm.seekp(0);
      memstm << "Finished removing value " << val << '.' << ends;
      sw_print(memstr);
#else      
      cout << "Value " << val << " removed.";
#endif
   }
}

/*
//////////////////////////////////////////////////
rb_node* redblack_tree::findhelp(rb_node* rt, const ELEM val)
//
// PURPOSE: Recursive function for finding a value in the tree.  Calls itself
//          on appropriate subtree until value found or current node is empty.
// PARAMETERS:
//    rt: Pointer to the current node
//    val: Value to be found
// RETURNS: Pointer to found value is now located
{
   rb_node* foundnode;
   if (rt == NULLNODE) {
      cout << "Value " << val << " not found in the tree.";
      return NULL;
   }
   else if (val == rt->value()) {
      cout << "Value " << val << " found in the tree.";
      return rt;
   }
   else if (val < rt->value()) return findhelp(rt->left, val, level + 1, LEFT);
   else return findhelp(rt->right, val, level + 1, RIGHT);
}
*/

//////////////////////////////////////////////////
void redblack_tree::printhelp(const rb_node* rt, const int level)
//
// PURPOSE: Recursive function for printing out the tree.  Prints a pre-order
//          traversal, with left children printed on same line as parent &
//          right children printed on following line.  Values are printed at
//          column 4 * level.
// PARAMETERS:
//    rt: Pointer to current node
//    level: Tree level of the current node */
{
   if (rt == NULLNODE) return;

   if (newline) for (int i = 0; i < level; i++) cout << "     ";   // Indent
   cout << setiosflags(ios::left) << setw(3) << rt->key
         << (rt->color == RED ? 'R': 'B') << ' ';
   newline = FALSE;

   printhelp(rt->right, level + 1);            // Print left subtree
   if (!newline) { cout << '\n'; newline = TRUE; } // Move down a line
   printhelp(rt->left, level + 1);       // Print right subtree
}

//////////////////////////////////////////////////
void redblack_tree::clearhelp(rb_node* rt)
//
// PURPOSE: Recursive function for clearing the tree.
// PARAMETERS:
//    rt: Pointer to current current node to be deleted */
{
   if (rt == NULLNODE) return;
   clearhelp(rt->left);
   clearhelp(rt->right);
   delete rt;
}

//////////////////////////////////////////////////
rb_node* redblack_tree::findmin(rb_node* rt)
//
// PURPOSE: Find the minimum value in a given subtree
// PARAMTERS:
//    rt: Pointer to root of subtree
{
   if (rt == NULLNODE) return NULLNODE;
   else if (rt->left != NULLNODE) return findmin(rt->left);
   else return rt;
}

