/*****************************************************************************
MODULE: grlist.cxx
AUTHOR: Cliff Shaffer/Jeff Nielsen
DATE: March 1996
DESCRIPTION: Implementation for graph: Adjacency list.  Mostly
             Dr. Shaffer's code  with some enhancements by Jeff Nielsen.
*****************************************************************************/
#include <stdlib.h>
#include <iostream.h>
#include <strstream.h>
#include <ctype.h>
#include <assert.h>
#include "grlist.h"

#define BUFLEN 80

int Graph::n()              // Number of vertices for graph
  { return numVertex; }

int Graph::e()              // Number of edges for graph
  { return numEdge; }

Edge Graph::first(int v)    // Get the first edge for a vertex
  { return list[v]; }

bool Graph::isEdge(Edge w)  // TRUE if this is an edge
  { return w != NULL; }

bool Graph::isEdge(int v1, int v2) {
   if (getEdge(v1,v2)) return TRUE;
   else return FALSE;
}

Edge Graph::getEdge(int v1, int v2) {
   Edge e;
   for (e = list[v1]; e != NULL && e->v2 != v2; e = e->next);
   return e;
   // Will return NULL if not found
}

Edge Graph::next(Edge w) {  // Get next edge for a vertex
  if (w == NULL) return NULL;
  else return w->next;
}

int Graph::v1(Edge w)       // Return vertex edge comes from
  { return w->v1; }

int Graph::v2(Edge w)       // Return vertex edge goes to
  { return w->v2; }

int Graph::weight(int i, int j) { // Return weight of edge
   Edge e;
   if ((e = getEdge(i,j)) != NULL) return e->weight;
   else return INFINITY;
}

int Graph::weight(Edge w) { // Return weight of edge
  if (w == NULL) return INFINITY;
  else return w->weight;
}

bool Graph::change_weight(int v1, int v2, int weight) {
   Edge e;
   if ((e = getEdge(v1,v2)) != NULL) {    // Make sure edge exists
      e->weight = weight;
      return TRUE;
   }
   else return FALSE;
}

bool Graph::add_edge(int v1, int v2, int weight) {
   Edge e;
   if (isEdge(v1,v2)) return FALSE;	  // Make sure not already an edge
   if ((e = list[v1]) == NULL || v2 < e->v2)
      list[v1] = new EdgeLink(v1, v2, weight, e);
   else {
      while (e->next != NULL && e->next->v2 < v2) e = e->next;
      e->next = new EdgeLink(v1, v2, weight, e->next);
   }
   numEdge++;
   return TRUE;
}

bool Graph::remove_edge(int v1, int v2) {
   Edge e, temp;
   if (!isEdge(v1, v2)) return FALSE;           // See if edge exists
   else if (list[v1]->v2 == v2) {
      temp = list[v1];
      list[v1] = list[v1]->next;
   }
   else {
      for (e = list[v1]; e->next->v2 != v2; e = e->next); //It's there
      temp = e->next;
      e->next = e->next->next;
   }
   delete temp;
   numEdge--;
   return TRUE;
}

Graph::Graph() {            // Default constructor
   init();
   invalid = FALSE;
}


Graph::Graph(istream& is) {
  char buffer[BUFLEN+1]; // Line buffer for reading
  bool undirected;  // TRUE if graph is undirected, FALSE if directed
  int i;
  int v1, v2, weight;
  Edge curr;

  init();        // In case input bad.
  is.getline(buffer, BUFLEN);
  numVertex = atoi(buffer);
  is.getline(buffer, BUFLEN);
  if (buffer[0] == 'U')
    undirected = TRUE;
  else if (buffer[0] == 'D')
    undirected = FALSE;
  else {
    cout << "Bad graph type: |" << buffer << "|\n";
    return;
  }

  // Create space
  mark = new bool[n()];
  list = new Edge [n()];
  pred = new int[n()];

  for (i=0; i< n(); i++) // Initialize matrix
    list[i] = NULL;

  // Read in edges
  numEdge = 0;
  while (true) {
	is.getline(buffer,BUFLEN);
    if (is.gcount() == 0) break;
    v1 = atoi(buffer);
    i = 0;
    while (isdigit(buffer[i])) i++;
    while (buffer[i] == ' ') i++;
    v2 = atoi(&buffer[i]);
    while (isdigit(buffer[i])) i++;
    if (buffer[i] == ' ') { // There is a weight
      while (buffer[i] == ' ') i++;
      weight = atoi(&buffer[i]);
    }
    else weight = 1;
    // Put the edge on the list
    numEdge++;
    if (((curr = list[v1]) == NULL) || (v2 < curr->v2))
      list[v1] = new EdgeLink(v1, v2, weight, curr);
    else {
      while ((curr->next != NULL) && (curr->next->v2 < v2))
         curr = curr-> next;
      curr->next = new EdgeLink(v1, v2, weight, curr->next);
    }
    if (undirected) {// Put in edge in other direction
      numEdge++;
      if (((curr = list[v2]) == NULL) || (v1 < curr->v2))
        list[v2] = new EdgeLink(v2, v1, weight, curr);
      else {
        while ((curr->next != NULL) && (curr->next->v2 < v1))
           curr = curr-> next;
        curr->next = new EdgeLink(v2, v1, weight, curr->next);
      }
    }
  }
  invalid = FALSE;
}

Graph::Graph(const Graph& other) {     // Copy constructor
   numVertex = other.numVertex;
   numEdge = other.numEdge;
   list = new Edge[numVertex];
   mark = new bool[numVertex];
   pred = new int[numVertex];

   // Make its own copy of the edge list
   Edge e, newe;
   for (int v = 0; v < numVertex; v++)
      if ((e = other.list[v]) == NULL) list[v] = NULL;
      else {
         newe = list[v] = new EdgeLink(*(other.list[v]));
         while (e->next != NULL) {
            e = e->next;
            newe->next = new EdgeLink(*e);
            newe = newe->next;
         }
      }
}

Graph::~Graph() {           // Destructor: return allocated space
   if (mark != NULL) delete [] mark;
   if (pred != NULL) delete [] pred;
   if (list != NULL) {
      for (int v = 0; v < numVertex; v++)
         while (list[v] != NULL) {
            Edge temp = list[v];
            list[v] = list[v]->next;
            delete temp;
         }
      delete [] list;
   }
}

