/*****************************************************************************
MODULE: vertcover.cpp
PROJECT: vertcover
AUTHOR: Jeff Nielsen
DATE: April 1996
DESCRIPTION: An implementation of the APPROX-VERTEX-COVER algorithm,
             annotated with SAIL function calls.
*****************************************************************************/
#include <fstream.h>
#include "stdjeff.h"
#include "grlist.h"
#ifndef SWAN
#define SWAN
#endif

// Swan functions -------------------------------------------------------------
#ifdef SWAN
#include <strstream.h>
#include <ctype.h>
#include "sail.h"

GRAPHID maingraph = GRAPHID(100);
char memstr[300];
strstream memstm(memstr, 300, ios::in | ios::out);  // Global in-memory stream

//////////////////////////////////////////////////
inline void make_swangraph(GRAPHID gid, NODETYPE nt, bool dir_undir,
            LAYOUTCOMPTYPE lctype)
//
// PURPOSE: Macro for making SWAN graphs
// PARAMETERS:
//    gid: The graphid you want for this graph
//    nt: The type of node you want the graph to have
//    dir_undir: DIRECTED or UNDIRECTED graph
//    lctype: The default layout component you want this graph to have.
{
   gid = sw_newgraph(gid, nt, dir_undir);
   sw_newlc(gid, LC_ID(1), lctype);
   sw_setcurlc(gid, LC_ID(1));
}

//////////////////////////////////////////////////
void drawgraph(Graph& g)
//
// PURPOSE: Draws the SWAN graph in the SWAN window
// PARAMETER;
//    g: The actual directed graph used in the program
{
   int v;
   for (v = 0; v < g.n(); v++) {
      memstm.seekp(0);
      memstm << v << ends;
      sw_insertnode(maingraph, NODEID(v), memstr);
   }
   for (v = 0; v < g.n(); v++)
      for (Edge w = g.first(v); g.isEdge(w); w = g.next(w) )
    sw_insertedge(maingraph, NODEID(w->v1), NODEID(w->v2), " ");

   sw_displayallgraphs();
}

//////////////////////////////////////////////////
void get_user_input(bool first, NODEID *node1 =NULL, NODEID *node2 = NULL)
//
// PURPOSE: Requests and stores user input
// PARAMETERS:
//    first: Whether this is the first time the function has been called
//    node1: Place to store the first vertex chosen by user
//    node2: Place to store the second vertex chosen by user (i.e., they chose
//	     an edge).
{
   static char cmd;
   if (first) do {
      sw_print("Enter 'e' to pick edges or 'v' to pick vertices.");
      sw_getstr(memstr);
      memstm.seekg(0);
      memstm >> cmd;
      if (tolower(cmd) != 'e' && tolower(cmd) != 'v') cmd = 0;
   } while (!cmd);
   
   else switch (tolower(cmd)) {
      case 'e':
	 sw_print("Please pick one of the edges remaining in E'.");
	 sw_pickedge(node1, node2);
	 break;
      case 'v':
	 sw_print("Please pick a vertex not in the cover.");
	 sw_picknode(node1);
	 break;
   }
}

#endif

// Vertex cover functions -----------------------------------------------------

//////////////////////////////////////////////////
void remove_incidents(Graph &g, int node)
//
// PURPOSE: Removes all of the edges incident on a vertex from the graph
// PARAMETERS:
//    g: The graph
//    node: The vertex from which you want to remove incident edges
{
   int v1, v2;
   Edge e = g.first(node);
   while (g.isEdge(e)) {
      v1 = e->v1;
      v2 = e->v2;
      e = g.next(e);
      g.remove_edge(v1, v2);
      g.remove_edge(v2, v1);
#ifdef SWAN
      // Color edges removed from E'
      sw_setedgeattr(maingraph, NODEID(v1), NODEID(v2),
		     GEDGECOLOR, LCYAN);
      sw_setedgeattr(maingraph, NODEID(v2), NODEID(v1),
		     GEDGECOLOR, LCYAN);
#endif
   }
}

//////////////////////////////////////////////////
void approx_vertcover(Graph &g)
//
// PURPOSE: This is the APPROX-VERTEX-COVER algorithm from Cormen, et al p.967
// PARAMETERS:
//    g: The graph
{
   int v, v1, v2;
   for (v = 0; v < g.n(); v++) g.mark[v] = FALSE; // Initialize mark array
   
   while (g.e() > 0) {	      // While the graph still has edges
#ifdef SWAN
      NODEID n1 = NODEID(UNDEF), n2 = NODEID(UNDEF);
      
      // Have user pick an edge or a vertex
      get_user_input(FALSE, &n1, &n2);
      v1 = int(n1);
      v2 = (n2 == NODEID(UNDEF)) ? 0 : int(n2);
      // Check their selection to make sure it's valid
      if (g.mark[v1] || (v2 && g.mark[v2])) {
	 if (v2) sw_print("ERROR: Edge no longer exists in E'.  Press STEP to "
			  "continue.");
	 else sw_print("ERROR: Vertex already in the cover.  Press STEP to "
		       "continue.");
	 sw_break();
	 continue;
      }
#else
      // Algorithm picks the first edge left in the graph
      for (v = 0; g.first(v) == NULL; v++);
      v1 = g.first(v)->v1;
      v2 = g.first(v)->v2;
#endif
      g.mark[v1] = TRUE;	 // Mark vertex as in the cover
      remove_incidents(g, v1);
#ifdef SWAN
      sw_setnodeattr(maingraph, n1, GNODEFILLED, TRUE);
#endif
      if (v2) {			 // Edge was chosen, rather than a vertex
	 g.mark[v2] = TRUE;	 // Mark vertex as in the cover
	 remove_incidents(g, v2);
#ifdef SWAN
	 sw_setnodeattr(maingraph, n2, GNODEFILLED, TRUE);
	 sw_setedgeattr(maingraph, n1, n2, GEDGELINETH, THICKLINE);
	 sw_setedgeattr(maingraph, n2, n1, GEDGELINETH, THICKLINE);
#endif
      }	    // End if (v2)
   }	    // End while (g.e() > 0)
}

//////////////////////////////////////////////////
void put_markedset(Graph &g, ostream& os)
//
// PURPOSE: Output the set of marked vertices
// PARAMETERS:
//    g: The graph
//    os: The output stream
{
   for (int v = 0; v < g.n(); v++)
      if (g.mark[v] == TRUE) os << v << "  ";
}


#ifdef SWAN
//////////////////////////////////////////////////
void swanmain(Graph &thegraph)
//
// PURPOSE: Main function for Swan visualization.  Makes it possible to run
//	    the visualization repeatedly
// PARAMETER:
//    thegraph: The graph to use for the visualization
{
   sw_disablebuttons(RUN);
   sw_step();
   make_swangraph(maingraph, CIRCLE, UNDIRECTED, CIRCLENET);
   sw_setgraphattr(maingraph, GEDGELABEL, FALSE);
   drawgraph(thegraph);
   get_user_input(TRUE);

   approx_vertcover(thegraph);	       // Find the vertex cover
   memstm.seekp(0);
   memstm << "This is a vertex cover: ";
   put_markedset(thegraph, memstm);
   memstm << ends;
   sw_print(memstr);
   sw_break();
   sw_deletegraph(maingraph);
}
#endif
   
//////////////////////////////////////////////////
int GS_MAIN(GS_MAIN_ARGS_F)
//
// PURPOSE: Driver for program
// PARAMETERS: The standard ones
{
   char    temp_str[80] ;
	
#ifdef SWAN
   sw_init(GS_MAIN_ARGS_A);
   sw_step();

   sw_print("What file would you like to use?") ;
   sw_getstr(temp_str) ;
   ifstream fs(temp_str);
   if (!fs) { sw_print("Couldn't open file "); exit(-1); }
   Graph maingraph(fs);    // Create graph from input
   if (!maingraph) { cout << "Unable to create graph.\n"; exit(-1);  }

   Graph *copygraph;
   while (TRUE) {
      copygraph = new Graph(maingraph);	     // All edges deleted each time
      swanmain(*copygraph);
      sw_print("Press RUN or STEP to run the visualization again.");
      sw_enablebuttons(RUN);
      sw_wait(RUN | STEP);
      delete copygraph;
   }
   sw_quit();
#else
   approx_vertcover(maingraph);	       // Find the vertex cover
   cout << "This is a vertex cover for the graph: ";
   put_markedset(maingraph, cout);
   cout << endl;
#endif
   return 0;
}

