/***********************************************************
 * circle.c -- a layout algorithm for undirected graphs    *
 * Jun Yang, November 17, 1994				   *
 ***********************************************************/

/*--------------- Header Files ---------------*/

#include <stdio.h>
#include <math.h>
#include "swan.h"
#include "laycn.h"

/**********************************************************
 * NOTE: The following macro to determine if two numbers  *
 * have the same sign, is for 2's complement number       *
 * representation. It will need to be modidfied for other *
 * number systems.					  *
 **********************************************************/

#define SAME_SIGN(a, b) \
	(((long)((unsigned long)a ^ (unsigned long)b)) >= 0)

//-------------------
//  CN_layout -- the main function for Circle-Net layout algorithm
//___________________

bool CN_layout(BasicGraph* g_ptr)
{
	BasicLayoutComp		*temp_lc ;
	BasicLayoutComp		*root_lc ;
	LCNBLIST		*temp_lclist ;

	if (g_ptr == NULL) {
	    _sw_errno = SWE_NOGRAPH ;
	    return FALSE ;
	}

	if (g_ptr->getNodeList()->getNodeCount() > NUMVERTEX) {
	    _sw_errno = SWE_MORENODES ;
	    return FALSE ;
	}

	if (g_ptr->getNodeList()->getNodeCount() == 0)
	    return TRUE ;

	root_lc = g_ptr->getRootLC() ;
	temp_lclist = root_lc->getChLCList() ;
	if (!temp_lclist->getFirstNode(temp_lc))
	    return FALSE ;

	if (temp_lc->getType() != CIRCLENET)
	    return FALSE ;

	if (!_cal_CN_box(g_ptr, temp_lc))
	    return FALSE ;
	root_lc->setBoxSize(temp_lc->getBoxSize()) ;

	if (!_cal_CN_pos(g_ptr, temp_lc))
	    return FALSE ;

	return TRUE ; 
} /* End of List_layout */

//-------------------
//  _cal_CN_box -- find the bounding box of the circle net
//___________________

bool _cal_CN_box(BasicGraph* g_ptr, BasicLayoutComp* lc_ptr) 
{
	int		Radius ;
	int		temp_r = MINEDGELEN ;
	unsigned int	n_node = lc_ptr->getChNodeList()->getNodeCount();
	BOXSIZE		size1 ;
	NODEINFO	*temp_node_info = lc_ptr->getNodeInfo() ;
	EDGEINFO	*temp_edge_info = lc_ptr->getEdgeInfo() ;

	switch (n_node) {
	case 0 :	// no node 
	    size1.w = size1.h = 0 ;
	    break ;

	case 1:		// only one node in the LC 	
	    size1 = temp_node_info->size ;
	    break ;

	default:
	    if (temp_r < temp_node_info->size.w)
		temp_r = temp_node_info->size.w ;
	    Radius = (int)((double)temp_r / sin(M_PI/n_node)) ;

	    // make Radius at least four times longer than node size
	    if (Radius < 4 * temp_node_info->size.w)
		Radius = 4 * temp_node_info->size.w ;
	    size1.w = 2 * (Radius + temp_node_info->size.w/2) + 1 ;
	    size1.h = 2 * (Radius + temp_node_info->size.h/2) + 1 ;
	    break ;
	}
	lc_ptr->setBoxSize(size1) ;
	return TRUE ;
} /* End of _cal_CN_box */

//-------------------
//  _cal_CN_pos -- calculate node positions on the circle
//___________________

bool _cal_CN_pos(BasicGraph* g_ptr, BasicLayoutComp* lc_ptr) 
{
        int             n, Radius ;
        double          alpha, delta_alpha ;
	bool		ok ;
        BasicNode       *cur_node ;
        LOCATION        pos1, cc ;
	NODELIST	*temp_nodelist = lc_ptr->getChNodeList() ;
	BOXSIZE		temp_node_size = lc_ptr->getNodeInfo()->size ;
 
        if ((n = temp_nodelist->getNodeCount()) == 0)
            return TRUE ;

        Radius = (lc_ptr->getBoxSize().w-1-temp_node_size.w)/2 ;
        delta_alpha = M_PI*2/n ;
        alpha = M_PI_2 ;
        cc.x = cc.y = Radius + 1 ;
        ok = temp_nodelist->getFirstNode(cur_node) ;
        while (ok) {
            pos1.x = (int)(cc.x + Radius * cos(alpha)) -
                     temp_node_size.w/2 ;
            pos1.y = (int)(cc.y + Radius * sin(alpha)) -
                     temp_node_size.h/2 ;
            cur_node->setRPos(pos1) ;
            alpha += delta_alpha ;
            ok = temp_nodelist->getNextNode(cur_node) ;
        }
	return TRUE ;
} /* End of _cal_CN_pos */

//-------------------
//  _CN_edge_ppath -- calculate physical path of edges in the circle
//___________________

bool _CN_edge_ppath(BasicGraph* g_ptr)
{
        bool         ok ;
        BasicEdge*      temp_edge ;
        BasicNode       *s_node, *e_node ;
        EDGELIST*       temp_edgelist ;
        CURVE           *ppath ;
        LOCATION        spos, epos, pos1, pos2 ;
	BasicLayoutComp *lc1, *lc2 ;
  
        temp_edgelist = g_ptr->getEdgeList() ;
        ok = temp_edgelist->getFirstNode(temp_edge) ;
        while (ok) {
            s_node = temp_edge->getSNode() ;
            e_node = temp_edge->getENode() ;
            lc1 = s_node->getLayoutComp() ;
            lc2 = e_node->getLayoutComp() ;
            ppath = temp_edge->getPPath() ;
            ppath->clear() ;
	    spos = s_node->getPPos() ;
	    epos = e_node->getPPos() ;
	    if (spos != UnsetPos && epos!= UnsetPos) {
	        if (lc1 == lc2) {
		    if (temp_edge->_intersections(spos, 
		            s_node->getType(), s_node->getBoxSize(),
		            epos, e_node->getType(),
		            e_node->getBoxSize(), pos1, pos2)) {
                        ppath->insert(pos1) ;
                        ppath->insert(pos2) ;
		    }
	        }
	        else 
		    return FALSE ;
	    }
            ok = temp_edgelist->getNextNode(temp_edge) ;
	}
	if (g_ptr->getGraphInfo()->fShowLabel)
	// If statement added by Jeff Nielsen 3/16/96 per instructions from Jun
	   _CN_edgelabel_pos(g_ptr) ;
	return TRUE ;
} /* End of _CN_edge_ppath */

//-------------------
//  _CN_edgelabel_pos -- calculate positions of edge labels
//___________________

bool _CN_edgelabel_pos(BasicGraph* g_ptr) 
{
        bool         ok, ok2, ok3 ;
        BasicEdge*      temp_edge ;
        BasicNode       *s_node, *e_node, *s_node2, *e_node2 ;
        LOCATION        spos, epos, spos2, epos2 ;
	LOCATION	pos1, pos2, pos3, pos4 ;
	NODELIST	*temp_nodelist ;
	NODENBLIST	*temp_nblist, *temp_nblist2 ;
	BasicNode	*temp_node, *cur_node ;
	BasicNode	**node_arr ;
	double		*alpha_arr, temp_alpha, PI_2 ;
	struct _nb_node *temp_nb, *temp_nb2 ;
	int		num_node, i, j, node_index, Radius ;
	int		left_limit, right_limit, end_pos ;
	double		cur_t ;
	BOXSIZE		temp_size ;
	LOCATION	cc ;
	NODEINFO	*temp_node_info ;
	BasicLayoutComp	*temp_lc ;
	BasicLayoutComp	*root_lc ;
	LCNBLIST	*temp_lclist ;
	EDGEINFO	*temp_info ;

	SortedDoubleList<double>  *temp_ipt_list = new
					SortedDoubleList<double> ;

	/*  sort the nodes according to the angles they incident
  	    with the center of the circle net.
	    find the center first */

	root_lc = g_ptr->getRootLC() ;
	temp_lclist = root_lc->getChLCList() ;
	if (!temp_lclist->getFirstNode(temp_lc))
	    return FALSE ;
	temp_size = temp_lc->getBoxSize() ;
	temp_node_info = temp_lc->getNodeInfo() ;
        Radius = (temp_size.w - 1) / 2 - temp_node_info->size.w/2 ;
        cc.x = cc.y = Radius + 1 - temp_node_info->size.w/2 ;
	
	// initial g_index
	temp_nodelist = g_ptr->getNodeList() ;
	num_node = temp_nodelist->getNodeCount() ;
	switch (num_node) {
	case 0 :
	case 1 :
	    return TRUE ;

	default:
	    PI_2 = M_PI * 2 ;
	    node_arr = new BasicNode*[num_node] ;
	    alpha_arr = new double[num_node] ;
	    i = 0 ;
	    ok = temp_nodelist->getFirstNode(temp_node) ;
	    while (ok) {
	        temp_node->g_index = i ;
	        node_arr[i] = temp_node ;
		pos1 = temp_node->getRPos() ;
		if (pos1.x > cc.x) {
		    temp_alpha = atan((double)(pos1.y - cc.y)/
			          (double)(pos1.x - cc.x)) ;
		    if (temp_alpha < 0.0)  // fourth quadrant
		        temp_alpha += PI_2 ;
		}
		else if (pos1.x < cc.x)  // second or third quadrant
		    temp_alpha = atan((double)(pos1.y - cc.y)/
			          (double)(pos1.x - cc.x)) + M_PI ;
		else 
		    temp_alpha = (pos1.y > cc.y) ? M_PI_2 : M_PI_2 +
							    M_PI ;
		alpha_arr[i++] = temp_alpha ;
	        ok = temp_nodelist->getNextNode(temp_node) ;
	    }
	    // sort the node array according to its angle
	    for (i = 0 ; i < num_node-1 ; i++) 
		for (j = i+1 ; j < num_node ; j++) 
		    if (alpha_arr[i] > alpha_arr[j]) {
		  	temp_alpha = alpha_arr[i] ;
			alpha_arr[i] = alpha_arr[j] ;
			alpha_arr[j] = temp_alpha ;
			node_arr[i]->g_index = j ;
			node_arr[j]->g_index = i ;
			temp_node = node_arr[i] ;
			node_arr[i] = node_arr[j] ;
			node_arr[j] = temp_node ;
		    }
	    delete[] alpha_arr ;
	    break ;
	}

	// process intersections
	for (i = 0 ; i < num_node ; i++) {
	     temp_node = node_arr[i] ;
	     left_limit = temp_node->g_index ;
	     temp_nblist = temp_node->getNBList() ;
	     ok2 = temp_nblist->getFirstNode(temp_nb) ;
	     while (ok2) {
		s_node = temp_nb->edge_ptr->getSNode() ;
		e_node = temp_nb->edge_ptr->getENode() ;

		right_limit = (s_node == temp_node) ? e_node->g_index :
						      s_node->g_index ;
		if (right_limit > left_limit) { // processing
		    spos = s_node->getPPos() ;
		    epos = e_node->getPPos() ;

		    temp_edge->_intersections(spos, s_node->getType(),
		        s_node->getBoxSize(), epos, e_node->getType(),
		        e_node->getBoxSize(), pos1, pos2) ;

		    temp_ipt_list->clear() ;
		    for (node_index = left_limit+1; 
		         node_index < right_limit; node_index++) {
		     cur_node = node_arr[node_index] ;
		     temp_nblist2 = cur_node->getNBList() ;
		     ok3 = temp_nblist2->getFirstNode(temp_nb2) ;
		     while (ok3) {
			s_node2 = temp_nb2->edge_ptr->getSNode() ;
			e_node2 = temp_nb2->edge_ptr->getENode() ;
			spos2 = s_node2->getPPos() ;
			epos2 = e_node2->getPPos() ;
			end_pos = (s_node2 == cur_node) ?
				  e_node2->g_index : s_node2->g_index ;
			if (end_pos > right_limit || 
			    end_pos < left_limit) {

			    // find the intersection
		temp_edge->_intersections(spos2, s_node2->getType(),
		    s_node2->getBoxSize(), epos2, e_node2->getType(),
		    e_node2->getBoxSize(), pos3, pos4) ;

			    if (lines_intersect(pos1, pos2, pos3, pos4,
			      		        cur_t))
			        temp_ipt_list->insert(cur_t) ;
			}
		        ok3 = temp_nblist2->getNextNode(temp_nb2) ;
		     }
		  } 

		  _find_largest_segment(pos1, pos2, temp_ipt_list,
				pos1, pos2) ;
		  temp_info = temp_nb->edge_ptr->getAttr() ;
            	  temp_size.w = CURFONTWIDTH * 
				strlen(temp_nb->edge_ptr->getLabel()) ;
            	  temp_size.h = CURFONTHEIGHT ;
            	  _sw_segedgepos(pos1, pos2, temp_size, temp_info) ;
		} 
	        ok2 = temp_nblist->getNextNode(temp_nb) ;
	    }
	}
	delete[] node_arr ;
	delete temp_ipt_list ;
	return TRUE ;
} /* End of _CN_edgelabel_pos */

//-------------------
//  _find_largest_segment -- find the largest line segment 
//___________________

void _find_largest_segment(LOCATION s_pos, LOCATION e_pos, 
			   SortedDoubleList<double> *t_list,
                           LOCATION& pos1, LOCATION& pos2) 
{
	long		x1, y1, x2, y2 ;
	long		delta_x, delta_y ;
	long		max_len = 0, length ;
	double 		prev_t, cur_t ;
	double		lt1, lt2 ;
	bool		ok ;
	
	if (t_list->getNodeCount() == 0) {
	    pos1 = s_pos ;
	    pos2 = e_pos ;
	    return ;
	}

	prev_t = 1.0 ;
	t_list->insert(prev_t) ;
	delta_x = (long) (e_pos.x - s_pos.x) ;
	delta_y = (long) (e_pos.y - s_pos.y) ;
	prev_t = 0.0 ;
	x1 = (long)s_pos.x ;
	y1 = (long)s_pos.y ;
	ok = t_list->getFirstNode(cur_t) ;
	while (ok) {
	    x2 = (long) (s_pos.x + cur_t * delta_x) ;
	    y2 = (long) (s_pos.y + cur_t * delta_y) ;
	    length = (x2 - x1) * (x2 - x1) + 
		     (y2 - y1) * (y2 - y1) ;
	    if (length > max_len) {
		max_len = length ;
		lt1 = prev_t ;
		lt2 = cur_t ;
	    }
	    x1 = x2 ;
	    y1 = y2 ;
	    prev_t = cur_t ;
	    ok = t_list->getNextNode(cur_t) ;
	}
	pos1.x = (int) (s_pos.x + lt1 * delta_x) ;
	pos1.y = (int) (s_pos.y + lt1 * delta_y) ;
	pos2.x = (int) (s_pos.x + lt2 * delta_x) ;
	pos2.y = (int) (s_pos.y + lt2 * delta_y) ;
} /* End of _find_largest_seg */

//-------------------
//  lines_intersect -- calculate the intersection point of two lines
//___________________

bool lines_intersect(LOCATION pos1, LOCATION pos2, 
		    LOCATION pos3, LOCATION pos4, double& ip_t)  
{
	long a1, a2, b1, b2, c1, c2 ; /* Coefficients of line eqns. */
	long denom, offset, num ;     /* Intermediate values */
	long t_x ;
	long r1, r2 ;	
	double temp_t = 0.0 ;

	/* Compute a1, b1, c1, where line joining points 1 and 2
 	 * is "a1 x + b1 y + c1 = 0".
 	 */

	a1 = pos2.y - pos1.y ;
	b1 = pos1.x - pos2.x ;
	c1 = (long)pos2.x * (long)pos1.y - (long)pos1.x * (long)pos2.y ;

	r1 = a1 * (long)pos3.x + b1 * (long)pos3.y + c1 ;
	r2 = a1 * (long)pos4.x + b1 * (long)pos4.y + c1 ;

	/* Check signs of r1 and r2. If both point 3 and point 4 lie on
	 * the smae side of line 1, the line segments do not intersect.
 	 */

	if (r1 != 0 && r2 != 0 && SAME_SIGN(r1, r2))
	    return FALSE ;

	/* Compute a2, b2, c2 */

	a2 = pos4.y - pos3.y ;
	b2 = pos3.x - pos4.x ;
	c2 = (long)pos4.x * (long)pos3.y - (long)pos3.x * (long)pos4.y ;

	r1 = a1 * (long)pos1.x + b1 * (long)pos1.y + c1 ;
	r2 = a1 * (long)pos2.x + b1 * (long)pos2.y + c1 ;

	/* Check signs of r1 and r2. If both point 1 and point 2 lie on
	 * the smae side of second line segment, 
         * the line segments do not intersect.
 	 */

	if (r1 != 0 && r2 != 0 && SAME_SIGN(r1, r2))
	    return FALSE ;

	/* Line segments intersect: comput intersection point.
	 */

	denom = a1 * b2 - a2 * b1 ;
	if ( denom == 0 )
	    return FALSE ; // collinear
	offset = denom < 0 ? - denom / 2 : denom /2 ;

	/* The denom/2 is to get rounding instead of truncating. It
	 * is added or subtracted to the numerator, depending upon the
	 * sign of the numerator.
	 */

	if (pos2.x != pos1.x) {
	    num = b1 * c2 - b2 * c1 ;
	    t_x = ( num < 0 ? num - offset : num + offset) / denom ;
	    temp_t = (double) (t_x - pos1.x) / (double) (pos2.x - pos1.x) ;
	}
	else if (pos2.y != pos1.y) {
	    num = a2 * c1 - a1 * c2 ;
	    t_x = ( num < 0 ? num - offset : num + offset) / denom ;
	    temp_t = (double) (t_x - pos1.y) / (double) (pos2.y - pos1.y) ;
	}
	
	if (temp_t <= 0.0 || temp_t >= 1.0)
	    return FALSE ;
	ip_t = temp_t ;
	return TRUE ;
} /* End of line_intersect */

/***********************************************************
 * End of circle.c         		   	 	   *
 ***********************************************************/

