/***********************************************************
 * kknet.c --- a layout algorithm for undirected graphs    *
 * Jun Yang, November 17, 1994				   *
 ***********************************************************/

/*-------------- References ---------------*/
/*
   This version for layout of undirected graph is an 
   implementation of the KAMADA and KAWAI's algorithm.
   
   Reference:

   [1] T. Kamada and S. Kawai, An Algorithm for Drawing
       General Undirected Graphs, Information Processing 
       Letters, Vol.31, pp.7-15, April, 1989.

 */

/*------------- Header Files ---------------*/

#include <stdio.h>
#include <math.h>
#include "swan.h"
#include "laykk.h"
#include "layman.h"

#ifdef NUMVERTEX
#undef NUMVERTEX
#define NUMVERTEX	25
#endif

#define EDGELENGTH	50.0
#define HOOKCONST	1.0 
#define WINFINITY	10000
#define EPSILONE	0.1
#define XYMARGIN	20.0

/*------------ Types ---------------*/

typedef int intMATRIX[NUMVERTEX][NUMVERTEX] ;
typedef double doubleMATRIX[NUMVERTEX][NUMVERTEX] ;

struct PARTICLE {
	BasicNode* ptr ; 
	double Edelta ;
	double x, y ;	// the particle's position
	double Px, Py ; // the partial derivative on (x,y)
	double Pxx, Pxy, Pyy ; // the second order partial derivative
} ;

struct edge_ip {
	BasicEdge *ptr ;
	SortedDoubleList<double> *ipt_list ;
} ;

typedef DoubleList<edge_ip*> EDGEIPLIST ;

/*-------------- Globale Variables ---------------*/

intMATRIX adjm ;	// ajacecy matrix of the graph
doubleMATRIX lm ;	// idea distance matrix 
doubleMATRIX km ;	// constant matrix 

PARTICLE nodearr[NUMVERTEX] ;
int 	 num_vertex ;

double oldMaxdelta_x, oldMaxdelta_y ;

//-------------------
//  KK_layout -- the main function for the KK layout algorithm
//___________________

bool KK_layout(BasicGraph* g_ptr)
{
	if (g_ptr == NULL) {
	    _sw_errno = SWE_NOGRAPH ;
	    return FALSE ;
	}

	if ((*g_ptr).isDirected()) 
	    return FALSE ;

	if ((*(*g_ptr).getNodeList()).getNodeCount() > NUMVERTEX)
	    return FALSE ;

	if ((*(*g_ptr).getNodeList()).getNodeCount() == 0)
	    return TRUE ;

	prepareMatrices(g_ptr) ;
	initNodeArr() ;
	_KK_iteration() ;
	assignPos() ;
	_sw_normalizeRPos(g_ptr) ;
	_KK_calboxsize(g_ptr) ;
	return TRUE ; 

} /* end of KK_layout */

void prepareMatrices(BasicGraph* g_ptr)
{
	register int	i = 0, j, k ;
	NODELIST	*temp_list ;
	NODENBLIST	*temp_nblist ;
	BasicNode	*temp_node ;
	_nb_node	*temp_nb ;
	bool		ok, ok2 ;
	int		temp_index ;

	temp_list = (*g_ptr).getNodeList() ;
	i = 0 ;
	ok = (*temp_list).getFirstNode(temp_node) ;
	while (ok) {
	    temp_node->g_index = i++ ;
	    ok = (*temp_list).getNextNode(temp_node) ;
	}

	i = 0 ;
	ok = (*temp_list).getFirstNode(temp_node) ;
	while (ok) {
	    nodearr[i].ptr = temp_node ;
	    for (j = 0 ; j < NUMVERTEX ; j++)
		adjm[i][j] = WINFINITY ;
	    temp_nblist = (*temp_node).getNBList() ;
	    ok2 = (*temp_nblist).getFirstNode(temp_nb) ;
	    while (ok2) {
		temp_index = temp_nb->id->g_index ;
		adjm[i][temp_index] = 1 ;
		ok2 = (*temp_nblist).getNextNode(temp_nb) ;
	    }
	    i++ ;
	    ok = (*temp_list).getNextNode(temp_node) ;
	}
	num_vertex = i ;

	/*
	   compute all-shortest paths in the graph using
	   Floyd's All_Pairs_Shortest_Paths algorithm

	   Reference:
	   [1] U. Manber, Introduction to Algorithms -- A Creative
	       Approach, Addison-Wesley, Reading, MA. 1989.
	*/

	for (k = 0 ; k < num_vertex ; k++) 
	    for (i = 0 ; i < num_vertex ; i++) 
	        for (j = 0 ; j < num_vertex ; j++) 
		    if (adjm[i][k] + adjm[k][j] < adjm[i][j])
			adjm[i][j] = adjm[i][k] + adjm[k][j] ;

	/* prepare lm and km */

	for (i = 0 ; i < num_vertex ; i++)
	    for (j = 0 ; j < num_vertex ; j++) {
		lm[i][j] = EDGELENGTH * adjm[i][j] ;
		km[i][j] = HOOKCONST / (adjm[i][j] * adjm[i][j]) ;
	    }
} /* end of prepareMatrices */

void initNodeArr(void)
{
	register int	i ;
	double Radius ;
	double delta_alpha ;
	double alpha = 0.0 ;
	double cc_x, cc_y ;

	delta_alpha = M_PI * 2.0 / num_vertex ;
	Radius = EDGELENGTH / (2.0 * sin(delta_alpha * 0.5)) ;
	cc_x = cc_y = Radius + XYMARGIN ;

	for (i = 0 ; i < num_vertex ; i++) {
	    nodearr[i].x = cc_x + Radius * cos(alpha) ;
	    nodearr[i].y = cc_y - Radius * sin(alpha) ;
	    alpha += delta_alpha ;
	}

	for (i = 0 ; i < num_vertex ; i++) 
	    calDelta_E(i) ;
	
} /* end of initNodeArr */

void _KK_iteration(void)
{
	int	maxEdelta = getMaxEdelta() ;
	double	oldMaxdelta_x, oldMaxdelta_y ;
	double  delta_x, delta_y ;

	while (nodearr[maxEdelta].Edelta > EPSILONE) {
	    oldMaxdelta_x = nodearr[maxEdelta].x ;
	    oldMaxdelta_y = nodearr[maxEdelta].y ;

	    while (nodearr[maxEdelta].Edelta > EPSILONE) {
		calDelta_xy(maxEdelta, &delta_x, &delta_y) ;
		nodearr[maxEdelta].x += delta_x ;
		nodearr[maxEdelta].y += delta_y ;
		calDelta_E(maxEdelta) ;
	    }

	    modEdelta(maxEdelta, oldMaxdelta_x, oldMaxdelta_y) ;
	    maxEdelta = getMaxEdelta() ;
	}
} /* end of _KK_iteration */

void calDelta_E(int index)
{
	register int	i ;
	double	dx, dy, d1, d3 ;
	double	xm = nodearr[index].x ;
	double	ym = nodearr[index].y ;
	double  E_Px = 0.0 ;
	double  E_Py = 0.0 ;
	double  E_Pxx = 0.0 ;
	double  E_Pxy = 0.0 ;
	double  E_Pyy = 0.0 ;
	double  kmi, lmi ;

	for (i = 0 ; i < num_vertex ; i++) 
	    if (i != index) {
	        dx = xm - nodearr[i].x ;
	        dy = ym - nodearr[i].y ;
		kmi = km[index][i] ;
		lmi = lm[index][i] ;
	        d1 = sqrt(dx * dx + dy * dy) ;
		d3 = d1 * d1 * d1 ;

                E_Px += kmi * (dx - lmi * dx / d1) ;
                E_Py += kmi * (dy - lmi * dy / d1) ;
		E_Pxx += kmi * (1 - lmi * dy * dy / d3) ;
		E_Pxy += kmi * lmi * dx * dy / d3 ;
		E_Pyy += kmi * (1 - lmi * dx * dx / d3) ;
	    } 

	nodearr[index].Edelta = sqrt(E_Px * E_Px + E_Py * E_Py) ;	
	nodearr[index].Px = E_Px ;
	nodearr[index].Py = E_Py ;
	nodearr[index].Pxx = E_Pxx ;
	nodearr[index].Pxy = E_Pxy ;
	nodearr[index].Pyy = E_Pyy ;
} /* end of calDelta_E */
	
void modEdelta(int max_index, double old_x, double old_y)
{
	register int 	i ;
	double cur_x = nodearr[max_index].x ;
	double cur_y = nodearr[max_index].y ;
	double dx, dy, d1, d3, kmi, lmi ;
	double old_Px, old_Py, old_Pxx, old_Pxy, old_Pyy ;
	double new_Px, new_Py, new_Pxx, new_Pxy, new_Pyy ;

	for (i = 0 ; i < num_vertex ; i++) 
	    if (i != max_index) {
	        dx = nodearr[i].x - old_x ;
	        dy = nodearr[i].y - old_y ;
		kmi = km[i][max_index] ;
		lmi = lm[i][max_index] ;
	        d1 = sqrt(dx * dx + dy * dy) ;
		d3 = d1 * d1 * d1 ;
		old_Px = kmi * (dx - lmi * dx / d1) ;
		old_Py = kmi * (dy - lmi * dy / d1) ;
		old_Pxx = kmi * (1.0 - lmi * dy * dy / d3) ;
		old_Pxy = kmi * lmi * dx * dy / d3 ;
		old_Pyy = kmi * (1.0 - lmi * dx * dx / d3) ;
 
	        dx = nodearr[i].x - cur_x ;
	        dy = nodearr[i].y - cur_y ;
	        d1 = sqrt(dx * dx + dy * dy) ;
		d3 = d1 * d1 * d1 ;
		new_Px = kmi * (dx - lmi * dx / d1) ;
		new_Py = kmi * (dy - lmi * dy / d1) ;
		new_Pxx = kmi * (1.0 - lmi * dy * dy / d3) ;
		new_Pxy = kmi * lmi * dx * dy / d3 ;
		new_Pyy = kmi * (1.0 - lmi * dx * dx / d3) ;

		nodearr[i].Px = nodearr[i].Px - old_Px + new_Px ;
		nodearr[i].Py = nodearr[i].Py - old_Py + new_Py ;
		nodearr[i].Pxx = nodearr[i].Pxx - old_Pxx + new_Pxx ;
		nodearr[i].Pxy = nodearr[i].Pxy - old_Pxy + new_Pxy ;
		nodearr[i].Pyy = nodearr[i].Pyy - old_Pyy + new_Pyy ;
	
		nodearr[i].Edelta = sqrt (nodearr[i].Px * nodearr[i].Px
				 + nodearr[i].Py * nodearr[i].Py) ;

	    }
} /* end of modEdelta */
	
int getMaxEdelta(void)
{
	register int	i ;
	double 		temp_max = nodearr[0].Edelta ;
	int		temp_index = 0 ;
	
	for (i = 1 ; i < num_vertex ; i++) 
	    if (nodearr[i].Edelta > temp_max) {
		temp_max = nodearr[i].Edelta ;
		temp_index = i ;
	    }
	return temp_index ;
} /* end of getMaxEdelta */

void calDelta_xy(int index, double* delta_x, double* delta_y)
{
	double d ;

	d = nodearr[index].Pxx * nodearr[index].Pyy -
	    nodearr[index].Pxy * nodearr[index].Pxy ;

	*delta_x = ((-nodearr[index].Px) * nodearr[index].Pyy -
		    (-nodearr[index].Py) * nodearr[index].Pxy) / d ;

	*delta_y = (nodearr[index].Pxx * (-nodearr[index].Py) -
		    nodearr[index].Pxy * (-nodearr[index].Px)) / d ;
} /* end of calDelta_xy */
		    
void assignPos(void)
{
	register int 	i, j ;
	LOCATION	temp_pos ;
	double		min_x = nodearr[0].x ;
	double 		min_y = nodearr[0].y ;
	double		delta_x, delta_y, cm_x, cm_y ;
	int		sortarr[NUMVERTEX] ;
	struct nodeangle {
	    double	delta_x ;
	    double	delta_y ;
	    double	tan_alpha ;
	} anglearr[NUMVERTEX] ;
	int		temp_index ;
	double Radius ;
	double delta_alpha ;
	double alpha = 0.0 ;
	double cc_x, cc_y ;
	
	for (i = 1 ; i < num_vertex ; i++) {
	     if (nodearr[i].x < min_x) 
		 min_x = nodearr[i].x ;
	     if (nodearr[i].y < min_y) 
		 min_y = nodearr[i].y ;
	}
	if (min_x < XYMARGIN) {
	    delta_x = XYMARGIN - min_x ;
	    for (i = 0 ; i < num_vertex ; i++)
		nodearr[i].x += delta_x ;
	}
	if (min_y < XYMARGIN) {
	    delta_y = XYMARGIN - min_y ;
	    for (i = 0 ; i < num_vertex ; i++)
		nodearr[i].y += delta_y ;
	}
 
/*
	cm_x = cm_y = 0.0 ;
	for (i = 0 ; i < num_vertex ; i++) {
	    cm_x += nodearr[i].x ;
	    cm_y += nodearr[i].y ;
	}
	cm_x /= num_vertex ;
	cm_y /= num_vertex ;

	temp_index = -1 ;
	for (i = 0 ; i < num_vertex ; i++) {
	    anglearr[i].delta_x = nodearr[i].x - cm_x ;
	    anglearr[i].delta_y = nodearr[i].y - cm_y ;
	    anglearr[i].tan_alpha = anglearr[i].delta_y /
				    anglearr[i].delta_x ;

	    for (j = i - 1 ; j >= 0 ; j--) 
		if (anglearr[i].delta_x > 0) {
		    if (anglearr[sortarr[j]].delta_x > 0) {
		        if (anglearr[i].tan_alpha > 
			    anglearr[sortarr[j]].tan_alpha) {
			    temp_index = j ;
			    break ;
			}
		    }
		}
		else {
		    if (anglearr[sortarr[j]].delta_x < 0) {		    
		        if (anglearr[i].tan_alpha > 
			    anglearr[sortarr[j]].tan_alpha) {
			    temp_index = j ;
			    break ;
			}
		    }
		    else {
			temp_index = j ;
			break ;
		    }
		}	
	
	    for (j = i - 1 ; j > temp_index ; j--) 
		sortarr[j+1] = sortarr[j] ;

	    sortarr[temp_index+1] = i ;
	}
			    
	delta_alpha = M_PI * 2.0 / num_vertex ;
	Radius = EDGELENGTH / (2.0 * sin(delta_alpha * 0.5)) ;
	cc_x = cc_y = Radius + XYMARGIN ;

	for (i = 0 ; i < num_vertex ; i++) {
	    temp_pos.x = (int)(cc_x + Radius * cos(alpha)) ;
	    temp_pos.y = (int)(cc_y - Radius * sin(alpha)) ;
	    (*(nodearr[sortarr[i]].ptr)).setRPos(temp_pos) ;
	    alpha += delta_alpha ;
	}
*/
// followings are original positions
	for (i = 0 ; i < num_vertex ; i++) {
	    temp_pos.x = (int)nodearr[i].x ;
	    temp_pos.y = (int)nodearr[i].y ;
	    (*(nodearr[i].ptr)).setRPos(temp_pos) ;
	}

} /* end of assignPos */
 
bool _KK_calboxsize(BasicGraph* g_ptr)
{
	BasicLayoutComp	*temp_lc, *root_lc ;
	LCNBLIST        *temp_lclist ;
	BOXSIZE		temp_size, temp_node_size ;
	NODELIST	*temp_nodelist ;
	BasicNode	*temp_node ;
	LOCATION	temp_pos ;
	bool		ok ;
	
        root_lc = (*g_ptr).getRootLC() ;
        temp_lclist = (*root_lc).getChLCList() ;
        if (!(*temp_lclist).getFirstNode(temp_lc))
            return FALSE ;

	temp_size.w = temp_size.h = 0 ;
	temp_nodelist = (*g_ptr).getNodeList() ;
	ok = (*temp_nodelist).getFirstNode(temp_node) ;
	while (ok) {
	    temp_pos = (*temp_node).getRPos() ;
	    temp_node_size = ((*temp_node).getAttr())->size ;
	    if (temp_pos.x + temp_node_size.w > temp_size.w)
		temp_size.w = temp_pos.x + temp_node_size.w ;
	    if (temp_pos.y + temp_node_size.h > temp_size.h)
		temp_size.h = temp_pos.y + temp_node_size.h ;
	    ok = (*temp_nodelist).getNextNode(temp_node) ;
	}
	(*temp_lc).setBoxSize(temp_size) ;
	(*root_lc).setBoxSize(temp_size) ;
	return TRUE ;
}


bool _KK_edge_ppath(BasicGraph* g_ptr)
{
        bool         ok ;
        BasicEdge*      temp_edge ;
        BasicNode       *s_node, *e_node ;
        EDGELIST*       temp_edgelist ;
        CURVE           *ppath ;
        LOCATION        pos1, pos2, spos, epos ;
	BasicLayoutComp *lc1, *lc2 ;

        temp_edgelist = (*g_ptr).getEdgeList() ;
        ok = (*temp_edgelist).getFirstNode(temp_edge) ;
        while (ok) {
            s_node = (*temp_edge).getSNode() ;
            e_node = (*temp_edge).getENode() ;
            lc1 = (*s_node).getLayoutComp() ;
            lc2 = (*e_node).getLayoutComp() ;
            ppath = (*temp_edge).getPPath() ;
            (*ppath).clear() ;

	    spos = (*s_node).getPPos() ;
	    epos = (*e_node).getPPos() ;
	    if (spos != UnsetPos && epos != UnsetPos) {
            if (lc1 == lc2) {
                if ((*temp_edge)._intersections(spos,
                    (*s_node).getType(), (*s_node).getBoxSize(),
                    epos, (*e_node).getType(),
                    (*e_node).getBoxSize(), pos1, pos2)) {
                    (*ppath).insert(pos1) ;
                    (*ppath).insert(pos2) ;
                }
	    }
            else
                return FALSE ;
	    }
            ok = (*temp_edgelist).getNextNode(temp_edge) ;
        }
	_Manual_edgelabel_pos(g_ptr) ;
	// _KK_edgelabel_pos(g_ptr) ;
        return TRUE ;
}

bool _KK_edgelabel_pos(BasicGraph* g_ptr) 
{
        bool         ok ;
        BasicEdge*      temp_edge ;
        BasicNode       *s_node, *e_node ;
        EDGELIST*       temp_edgelist ;
        LOCATION        spos, epos, pos1, pos2 ;
	EDGEINFO	*temp_info ;
	BOXSIZE		temp_size ;
	LABEL		temp_label ;
  
        temp_edgelist = (*g_ptr).getEdgeList() ;
        ok = (*temp_edgelist).getFirstNode(temp_edge) ;
        while (ok) {
            s_node = (*temp_edge).getSNode() ;
            e_node = (*temp_edge).getENode() ;
            spos = (*s_node).getPPos() ;
            epos = (*e_node).getPPos() ;

	    if (spos != UnsetPos && epos != UnsetPos) {
            if (!(*temp_edge)._intersections((*s_node).getPPos(),
                (*s_node).getType(), (*s_node).getBoxSize(),
                (*e_node).getPPos(), (*e_node).getType(),
                (*e_node).getBoxSize(), pos1, pos2)) 
		return FALSE ;

	    temp_label = (*temp_edge).getLabel() ;
	    temp_size.w = CURFONTWIDTH * strlen(temp_label) ;
	    temp_size.h = CURFONTHEIGHT ;
	    temp_info = (*temp_edge).getAttr() ;
	    _sw_segedgepos(pos1, pos2, temp_size, temp_info) ;
	    } 
            ok = (*temp_edgelist).getNextNode(temp_edge) ;
	}
	return TRUE ;
}

/***********************************************************
 * End of kknet.c           				   *
 ***********************************************************/
